from datetime import datetime
import snappy
from test_base import is_symplectic, testing_string
from multiprocessing import Pool
import itertools
import random

start = 0
end = 1
scale = 500
test = "random"

print(f"[{datetime.now().strftime('%d-%m-%y %H:%M:%S')}]    Building test set")

manifolds = [random.randint(1, len(snappy.HTLinkExteriors)) for _ in range(scale * (end - start))]
manifolds_tri = [snappy.HTLinkExteriors[i] for i in manifolds]
manifolds_labels = [M.identify()[0] for M in manifolds_tri if len(M.identify()) > 0]


def process_manifold(i: int, output: bool = True):
    M = manifolds_tri[i]
    index = manifolds[i]
    label = manifolds_labels[i]

    # print(index)

    if index == 0:
        return True

    basis = M.symplectic_basis()
    result = is_symplectic(basis)

    if result:
        string = "Passed"
    else:
        string = "Failed"

    if output:
        with open(f"logs/links-{i // scale}.log", "a") as file:
            file.write(f"Testing: {str(index)} {(20 - len(str(index))) * ' '} {str(label)} {(40 - len(str(label))) * ' '} {string}\n")

    print(f"Testing: {str(index)} {(20 - len(str(index))) * ' '} {str(label)} {(40 - len(str(label))) * ' '} {string}")

    return result


def test_link_complements_pool():
    with open("logs/total.log", "a") as file:
        file.write(testing_string(scale * (end - start)))

    with Pool(maxtasksperchild=25) as pool:
        if test == "random":
            result = pool.imap(process_manifold, range(len(manifolds)))
        else:
            result = pool.imap(process_manifold, range(start * scale, end * scale))

        for _ in range(start, end):
            lst = list(itertools.islice(result, scale))

            time = datetime.now().strftime('%d-%m-%y %H:%M:%S')
            # print(f"[{time}]    Passed: {sum(lst)} / {len(lst)}")

            with open("logs/total.log", "a") as file:
                file.write(f"[{time}]    Passed: {sum(lst)} / {len(lst)}\n")


if __name__ == "__main__":
    test_link_complements_pool()
    # unittest.main()



import random
from datetime import datetime
import snappy
import unittest
import spherogram
from tqdm import tqdm

start = 0
end = 1
scale = 1000
num_tests = 5000
test = "random"

if len(snappy.HTLinkExteriors(crossings=15)) == 0:
    file_name = "small-db"
else:
    file_name = "large-db"


def is_symplectic(M):
    """
    Test if the matrix M is symplectic
    :param M: square matrix
    :return: true or false
    """
    n = M.dimensions()

    for i in range(n[0]):
        for j in range(i, n[1]):
            omega = abs(symplectic_form(M.data[i], M.data[j]))

            if i % 2 == 0 and j % 2 == 1 and j == i + 1:
                if omega != 2:
                    return False
            elif omega:
                return False

    return True


def symplectic_form(u, v):
    return sum([u[2 * i] * v[2 * i + 1] - u[2 * i + 1] * v[2 * i]
                for i in range(len(u) // 2)])


def save_manifold(index: int):
    M = snappy.HTLinkExteriors[index]
    M.save(f"CuspedCensusData/link-{index}.tri")


def process_manifold(index: int):
    M = snappy.HTLinkExteriors[index]
    label = M.identify()[0] if len(M.identify()) > 0 else ""

    if index == 0:
        return True

    print(index)

    basis = M.symplectic_basis()
    result = is_symplectic(basis)

    if result:
        string = "Passed"
    else:
        string = "Failed"

    print(f"Testing: {str(index)} {(20 - len(str(index))) * ' '} {str(label)} {(40 - len(str(label))) * ' '} {string}")

    return result


def random_link_exteriors(n: int, n_tet: int, n_cusps: int):
    for i in range(n):
        L = spherogram.random_link(n_tet, n_cusps, alternating=True)
        M = spherogram.Link.exterior(L)
        print(M.num_cusps())
        M.save(f"CuspedCensusData/link-{n_tet}-{n_cusps}-{i}.tri")


def generate_tests(output: bool):
    print(f"[{datetime.now().strftime('%d-%m-%y %H:%M:%S')}]  Generating symplectic basis tests")
    for _ in range(2000):
        index = random.randint(1, len(snappy.HTLinkExteriors) - 1)
        process_manifold(index)

        if output:
            with open(file_name, "a") as file:
                file.write(f"{index}\n")


def testing_string(num: int = 0):
    time = datetime.now().strftime('%d-%m-%y %H:%M:%S')

    if test == "random":
        return f"[{time}]    Testing ({test}): {num} manifolds\n"
    elif test == "sequence":
        return f"[{time}]    Testing ({test}): {scale * start} - {scale * end}\n"


def test_link_complements():
    """ Pick 'num_test' random manifolds from HTLinkExteriors and test symplectic_basis() """
    # with open(file_name, "r") as file:
    #     lst = file.readlines()

    # manifolds = list(set([int(x[:-1]) for x in lst]))

    # with open("logs/total.log", "a") as file:
    #     file.write(testing_string(len(manifolds)))
    print(testing_string(num_tests))

    # if test == "random":
    #     result = [process_manifold(i) for i in range(len(manifolds))]
    # else:
    #     result = [process_manifold(i) for i in range(scale * start, scale * end)]
    manifolds = [random.randint(1, len(snappy.HTLinkExteriors)) for _ in range(num_tests)]
    result = [process_manifold(i) for i in manifolds]

    # with open("logs/total.log", "a") as file:
    #     file.write(f"[{datetime.now().strftime('%d-%m-%y %H:%M:%S')}]    Passed: {sum(result)} / {len(result)}\n")
    print(f"[{datetime.now().strftime('%d-%m-%y %H:%M:%S')}]    Passed: {sum(result)} / {len(result)}")


class TestSymplecticBasis(unittest.TestCase):
    def test_knot_complements(self):
        i = 0
        for M in tqdm(snappy.CensusKnots, desc="Knots...", ncols=120):
            with self.subTest(i=i):
                # print(M.identify()[0])
                basis = M.symplectic_basis()
                self.assertTrue(is_symplectic(basis), str(M.identify()[0]))
                i += 1

    # @unittest.skip
    def test_link_complements(self):
        i = 0
        for M in tqdm(snappy.HTLinkExteriors[1:1000], desc="Links...", ncols=120):
            with self.subTest(i=i):
                # print(M.identify()[0])
                basis = M.symplectic_basis()
                self.assertTrue(is_symplectic(basis))
                i += 1

    @unittest.skip
    def test_random_links(self):
        iterations = 10

        for i in tqdm(range(iterations), desc="Random Links...", ncols=120):
            with self.subTest(i=i):
                L = spherogram.random_link(100, num_components=random.randint(3, 10), alternating=True)
                M = spherogram.Link.exterior(L)
                basis = M.symplectic_basis()
                self.assertTrue(is_symplectic(basis))


if __name__ == "__main__":
    # test_link_complements()
    # generate_tests(True)
    # unittest.main()

    save_manifold(116084)



from datetime import datetime
import snappy
from multiprocessing import Pool
import itertools
from test_base import is_symplectic, testing_string
import random

start = 0
end = 1
scale = 1000
test = "random"

if len(snappy.HTLinkExteriors(crossings=15)) == 0:
    file_name = "small-db"
else:
    file_name = "large-db"


def process_manifold(index: int, output: bool = True):
    if test == "random":
        index = random.randint(1, len(snappy.HTLinkExteriors) - 1)

    M = snappy.HTLinkExteriors[index]
    label = M.identify()[0] if len(M.identify()) > 0 else ""
    # print(label)

    if index == 0:
        return True

    basis = M.symplectic_basis()
    result = is_symplectic(basis)

    if result:
        string = "Passed"
    else:
        string = "Failed"

    if output:
        with open("logs/links-0.log", "a") as file:
            file.write(f"Testing: {str(index)} {(20 - len(str(index))) * ' '} {str(label)} {(40 - len(str(label))) * ' '} {string}\n")

    # print(f"Testing: {str(index)} {(20 - len(str(index))) * ' '} {str(label)} {(40 - len(str(label))) * ' '} {string}")

    return result


def test_link_complements_pool(manifolds):
    with open("logs/total.log", "a") as file:
        if test == "database":
            length = len(manifolds)
        else:
            length = scale * (end - start)

        file.write(testing_string(length))
        print(testing_string(length))

    with Pool(maxtasksperchild=25) as pool:
        if test == "database":
            result = pool.imap(process_manifold, manifolds)
        else:
            result = pool.imap(process_manifold, range(start * scale, end * scale))

        for _ in range(start, end):
            lst = list(itertools.islice(result, scale))

        # lst = list(result)
            time = datetime.now().strftime('%d-%m-%y %H:%M:%S')
            print(f"[{time}]    Passed: {sum(lst)} / {len(lst)}")

            with open("logs/total.log", "a") as file:
                file.write(f"[{time}]    Passed: {sum(lst)} / {len(lst)}\n")


if __name__ == "__main__":
    with open(file_name, "r") as file:
        lst = file.readlines()

    manifolds = list(set([int(x[:-1]) for x in lst]))
    test_link_complements_pool(manifolds)
    # test_link_complements()
    # generate_tests()
    # unittest.main()



"""

Cython file

"""

cdef extern from "triangulation.h":
    ctypedef struct c_Triangulation "Triangulation":
        pass

cdef extern from "SnapPea.h":
    void peripheral_curves(c_Triangulation *manifold)
    int *get_cusp_equation(c_Triangulation *manifold, int, int, int, int *)
    void free_cusp_equation(int *eqn)

cdef extern from "unix_file_io.h":
    c_Triangulation *read_triangulation_from_string(char *string)

cdef extern from "symplectic_basis.h":
    int **get_symplectic_basis(c_Triangulation *manifold, int *num_rows, int *num_cols, int log)
    void free_symplectic_basis(int **eqns, int num_rows)

def symplectic_basis(manifold, verify=False):
    """
    Extend the Neumann-Zagier Matrix to one which is symplectic (up to factors of 2)
    using oscillating curves. Verify parameter explicitly tests if the resulting matrix is symplectic.

    >>> M = Manifold("4_1")
    >>> M.symplectic_basis()
    [-1  0 -1 -1]
    [ 2  0 -2  0]
    [-2 -1 -2 -1]
    [ 2  0  0  0]

    <https://arxiv.org/abs/2208.06969>
    """
    def to_byte_str(s):
        return s.encode('utf-8') if type(s) != bytes else s

    def is_symplectic(M):
        """
        Test if the matrix M is symplectic
        :param M: square matrix
        :return: true or false
        """
        n = len(M)

        for i in range(n):
            for j in range(i, n):
                omega = abs(symplectic_form(M[i], M[j]))

                if i % 2 == 0 and j % 2 == 1 and j == i + 1:
                    if omega != 2:
                        return False
                elif omega:
                    return False

        return True

    def symplectic_form(u, v):
        return sum([u[2 * i] * v[2 * i + 1] - u[2 * i + 1] * v[2 * i] for i in range(len(u) // 2)])

    cdef int **c_eqns;
    cdef int **g_eqns;
    cdef int num_rows, num_cols, dual_rows;
    cdef int* eqn;
    cdef c_Triangulation *c_triangulation = NULL;

    string = manifold._to_string()
    b_string = to_byte_str(string)
    c_triangulation = read_triangulation_from_string(b_string)

    if c_triangulation is NULL:
        raise ValueError('The Triangulation is empty.')

    eqns = []

    peripheral_curves(c_triangulation)

    # Cusp Equations
    for i in range(manifold.num_cusps()):
        cusp_info = manifold.cusp_info(i)
        if cusp_info.is_complete:
            to_do = [(1,0), (0,1)]
        else:
            to_do = [cusp_info.filling]
        for (m, l) in to_do:
            eqn = get_cusp_equation(c_triangulation, i, int(m), int(l), &num_rows)
            eqns.append([eqn[j] for j in range(num_rows)])
            free_cusp_equation(eqn)

    # Dual Curve Equations
    g_eqns = get_symplectic_basis(c_triangulation, &dual_rows, &num_cols, 0)

    for i in range(dual_rows):
        eqns.append([g_eqns[i][j] for j in range(num_cols)])

    free_symplectic_basis(g_eqns, dual_rows)

    # Convert to Neumann Zagier Matrix
    rows = len(eqns)
    retval = [[eqns[i][3 * (j // 2) + j % 2] - eqns[i][3 * (j // 2) + 2] for j in range(rows)] for i in range(rows)]

    if verify:
        if is_symplectic(retval):
            print("Result is symplectic (up to factors of 2)")
        else:
            print("Warning: Result is not symplectic")

    return retval


import os
from distutils.core import setup
from distutils.extension import Extension
from Cython.Build import cythonize
from setuptools.glob import glob

base_code = glob(os.path.join('snappy_src/kernel_code', '*.c'))
symp_code = glob(os.path.join('symp_src', '*.c'))
symp_source_files = base_code + symp_code

symp_source_files.append("cython/symp_basis.pyx")

base_head = glob(os.path.join('snappy_src/headers', '*.h'))
symp_head = glob(os.path.join('symp_src', '*.h'))

symp_header_files = base_head + symp_head

symp_ext = Extension(
    name="symplectic_basis",
    sources=symp_source_files,
    library_dirs=["symp_src",
                  "snappy_src/kernel_code",
                  "snappy_src/headers"],
    include_dirs=["symp_src",
                  "snappy_src/kernel_code",
                  "snappy_src/headers"],
    language="c"
)

setup(
    name="symplectic-basis",
    headers=symp_header_files,
    ext_modules=cythonize(symp_ext, compiler_directives={'language_level': "3"})
)